import json
from collections import defaultdict
from io import BytesIO
from typing import Any, Dict, Tuple
from unittest import mock
from urllib.parse import urlencode

import pytest
from django.contrib.auth import get_user_model
from faker import Faker
from requests import Response

fake = Faker()


@pytest.fixture(autouse=True)
def enable_db_access_for_all_tests(db):
    pass


@pytest.fixture
def user_mei():
    user_cls = get_user_model()
    try:
        user = user_cls.objects.get(username="mei")
    except user_cls.DoesNotExist:
        user = user_cls()
        user.username = "mei"
        user.is_staff = True
        user.email = "mei@toto.ro"
        user.first_name = "Mei"
        user.last_name = "Kusakabe"
        user.is_superuser = True
        user.save()
    return user


@pytest.fixture
def random_userinfo():
    return {
        "username": f"{fake.user_name()}_{fake.pystr(4, 4)}",
        "is_staff": True,
        "email": fake.email(),
        "first_name": fake.first_name(),
        "last_name": fake.last_name(),
        "is_superuser": True,
    }


API_RESPONSE: Dict[Tuple[str, str], Tuple[int, Any, Any]] = {
    (
        "post",
        "https://gandi.v5/token~client_id=Mei&client_secret=T0t0r0&code=codecode&grant_type=authorization_code",
    ): (
        200,
        # Headers
        {},
        # Body
        {"access_token": "mey_accesstoken", "refresh_token": "freshmenthol"},
    ),
    (
        "post",
        "https://gandi.v5/token~client_id=Mei&client_secret=T0t0r0&code=badcode&grant_type=authorization_code",
    ): (
        400,
        # Headers
        {},
        # Body
        {"error": "invalid_token"},
    ),
    (
        "post",
        "https://gandi.v5/token~client_id=Mei&client_secret=T0t0r0&code=e500&grant_type=authorization_code",
    ): (
        500,
        # Headers
        {},
        # Body
        {"error": "server_error"},
    ),
    (
        "post",
        "https://gandi.v5/token~client_id=Mei&client_secret=T0t0r0&grant_type=refresh_token&refresh_token=freshmenthol",
    ): (
        200,
        # Headers
        {},
        # Body
        {"access_token": "def", "refresh_token": "freshmenthol"},
    ),
    (
        "post",
        "https://gandi.v5/token~client_id=Mei&client_secret=T0t0r0&grant_type=refresh_token&refresh_token=e409",
    ): (
        409,
        # Headers
        {},
        # Body
        {},
    ),
    (
        "post",
        "https://gandi.v5/token~client_id=Mei&client_secret=T0t0r0&grant_type=refresh_token&refresh_token=e500",
    ): (
        500,
        # Headers
        {},
        # Body
        {},
    ),
}


class DummyResponse(Response):
    """Represent a requests response."""

    def __init__(self, body="", status_code=200, headers=None):
        super(DummyResponse, self).__init__()
        self.raw = BytesIO(body.encode("utf-8"))
        self.status_code = status_code
        self.headers = headers or {}


class RequestsMock(mock.Mock):
    """A mock for request calls."""

    api_calls = defaultdict(list)

    def __init__(
        self,
        api_response,
        missing_status=503,
        missing_body="Internal Server Error",
    ):
        api_response = api_response

        def fake_api(method, url, **kwargs):
            get_params = sorted(kwargs.get("params", {}).items())
            if get_params:
                url += "?" + urlencode(get_params)
            post_params = sorted(kwargs.get("data", {}).items())
            if post_params:
                url += "~" + urlencode(post_params)

            if "TimeoutError" in url:
                raise TimeoutError("Boom")

            if "ConnectionError" in url:
                raise ConnectionError("Boom")

            if "RuntimeError" in url:
                raise RuntimeError("Sometime things fails")

            if (method, url) in api_response:
                status, headers, res = api_response[(method, url)]
                res = json.dumps(res)
            else:
                print(f"{method} {url} is missing returning missing results")
                status, headers, res = missing_status, None, missing_body
            RequestsMock.api_calls[(method, url)].append(kwargs)
            return DummyResponse(res, status, headers)

        super(RequestsMock, self).__init__(side_effect=fake_api)


@pytest.fixture()
def mock_oauth2():
    mock_req = mock.patch(
        "requests.Session.request",
        RequestsMock(API_RESPONSE),
    )
    mock_req.start()
    yield RequestsMock
    mock_req.stop()
    RequestsMock.api_calls.clear()


@pytest.fixture()
def state():
    state = fake.pystr(4, 4)
    with mock.patch("wagtail_oauth2.views.gen_state_name", return_value=state):
        yield state


@pytest.fixture()
def auth_code(mock_oauth2):
    return "codecode"



import pytest

from wagtail_oauth2.resources import Token


def test_token_by_authcode(mock_oauth2):
    tokens = Token.by_authcode("codecode")
    assert tokens == {
        "access_token": "mey_accesstoken",
        "refresh_token": "freshmenthol",
    }
    assert mock_oauth2.api_calls == {
        (
            "post",
            "https://gandi.v5/token~client_id=Mei&client_secret=T0t0r0&code=codecode&grant_type=authorization_code",
        ): [
            {
                "data": {
                    "client_id": "Mei",
                    "client_secret": "T0t0r0",
                    "code": "codecode",
                    "grant_type": "authorization_code",
                },
                "json": None,
                "timeout": 30,
                "verify": False,
            }
        ]
    }


@pytest.mark.parametrize("authcode", ["badcode", "e500", "TimeoutError", "ConnectionError", "RuntimeError"])
def test_token_by_authcode_unkown_user(authcode, mock_oauth2):
    tokens = Token.by_authcode(authcode)
    assert tokens == {}


def test_get_authenticated_url():
    url = Token.get_authenticated_url("http://my.cms", "abahoui")
    assert (
        url
        == "https://gandi.v5/authorize?client_id=Mei&redirect_uri=http%3A%2F%2Fmy.cms&response_type=code&state=abahoui"
    )


def test_by_refresh_token(mock_oauth2):
    tokens = Token.by_refresh_token("freshmenthol")
    assert tokens == {
        "access_token": "def",
        "refresh_token": "freshmenthol",
    }
    assert mock_oauth2.api_calls == {
        (
            "post",
            "https://gandi.v5/token~client_id=Mei&client_secret=T0t0r0&grant_type=refresh_token&refresh_token=freshmenthol",
        ): [
            {
                "data": {
                    "client_id": "Mei",
                    "client_secret": "T0t0r0",
                    "grant_type": "refresh_token",
                    "refresh_token": "freshmenthol",
                },
                "json": None,
                "timeout": 30,
                "verify": False,
            }
        ]
    }

@pytest.mark.parametrize("refresh_troken", ["e409", "e500", "TimeoutError", "ConnectionError", "RuntimeError"])
def test_by_refresh_token_4xx(refresh_troken, mock_oauth2):
    tokens = Token.by_refresh_token(refresh_troken)
    assert tokens == {}



import pkg_resources

try:
    __version__ = pkg_resources.get_distribution("wagtail_oauth2").version
except pkg_resources.DistributionNotFound:
    # Read the docs does not support poetry and cannot install the package
    __version__ = None



"""
Manage OAuth2.0 Resources.
"""
import logging

from urllib.parse import urlencode

import requests
from requests.exceptions import HTTPError, ConnectionError, Timeout

from .settings import get_setting

log = logging.getLogger(__name__)


class Token:
    """Retrieve tokens from the OAuth2.0 Resource server."""
    @classmethod
    def by_authcode(cls, auth_code):
        "Retrieve Tokens from an authorization code."
        try:
            token_url = get_setting("TOKEN_URL")
            log.info("Fetching token on %s/token" % token_url)
            response = requests.post(
                token_url,
                data={
                    "client_id": get_setting("CLIENT_ID"),
                    "client_secret": get_setting("CLIENT_SECRET"),
                    "grant_type": "authorization_code",
                    "code": auth_code,
                },
                verify=get_setting("VERIFY_CERTIFICATE", True),
                timeout=get_setting("TIMEOUT", 30),
            )
            response.raise_for_status()
            return response.json()

        except (ConnectionError, TimeoutError, Timeout) as exc:
            log.error("OAuth2 server is down: %s" % exc.__class__.__name__)
        except HTTPError as exc:
            if exc.response.status_code >= 500:
                log.error(
                    "OAuth2 server is not working properly? got %s"
                    % exc.response.status_code
                )
            else:
                log.error("Cannot retrieve token: %s" % exc.response.text)
        except Exception:
            log.exception(
                "Unexpected exception while retrieving token using "
                "authorization code"
            )
        return {}

    @classmethod
    def get_authenticated_url(cls, login_url, state):
        "Get the authorization url with its parameter."
        data = {
            "client_id": get_setting("CLIENT_ID"),
            "redirect_uri": login_url,
            "response_type": "code",
            "state": state,
        }
        return "{}?{}".format(
            get_setting("AUTH_URL"), urlencode(data, doseq=True)
        )

    @classmethod
    def by_refresh_token(cls, refresh_token):
        "Retrieve Tokens from a refresh tokend."
        try:
            token_url = get_setting("TOKEN_URL")
            log.info("Refreshing token on %s/token" % token_url)
            response = requests.post(
                token_url,
                data={
                    "client_id": get_setting("CLIENT_ID"),
                    "client_secret": get_setting("CLIENT_SECRET"),
                    "grant_type": "refresh_token",
                    "refresh_token": refresh_token,
                },
                verify=get_setting("VERIFY_CERTIFICATE", True),
                timeout=get_setting("TIMEOUT", 30),
            )
            response.raise_for_status()
            tokens = response.json()
            return tokens
        except (ConnectionError, TimeoutError, Timeout) as exc:
            log.error("OAuth2 server is down: %s" % exc.__class__.__name__)
        except HTTPError as exc:
            if 400 <= exc.response.status_code < 500:
                log.warning(
                    "OAuth2 server does not refresh the token (%s: %s), "
                    "force disconnect",
                    exc.response.status_code,
                    exc.response.text,
                )
            elif 500 <=exc.response.status_code < 600:
                log.error(
                    "OAuth2 server is not working properly? got %s",
                    exc.response.status_code,
                )
        except Exception:
            log.exception(
                "Unexpected exception while retrieving token using refresh token"
            )
        return {}



import re

from django.contrib.auth.models import User

from wagtail_oauth2.views import gen_state_name, get_cookie_name, get_user_from_userinfo


def test_gen_state_name():
    state1 = gen_state_name()
    state2 = gen_state_name()
    assert state1 != state2
    assert re.match("[0-9a-z]{10}", state1)
    assert re.match("[0-9a-z]{10}", state2)


def test_get_cookie_name():
    return get_cookie_name("plop") == "oauth2.plop"


def test_get_user_from_userinfo_exists(user_mei):
    user = get_user_from_userinfo({"username": "mei"})
    assert user is not None
    assert isinstance(user, User)
    assert user == user_mei


def test_get_user_from_userinfo_does_not_exists(random_userinfo):
    user = get_user_from_userinfo(random_userinfo)
    assert user is not None
    assert isinstance(user, User)
    assert user.username == random_userinfo["username"]
    assert user.email == random_userinfo["email"]
    assert user.first_name == random_userinfo["first_name"]
    assert user.last_name == random_userinfo["last_name"]


def test_view_get(client, state):
    resp = client.get("/admin/login/")
    assert resp.status_code == 302
    redir = "https%3A%2F%2Ftestserver%2Fadmin%2Flogin%2F"
    assert (
        resp.headers["location"]
        == f"https://gandi.v5/authorize?client_id=Mei&redirect_uri={redir}&response_type=code&state={state}"
    )
    state_cookie = resp.cookies[get_cookie_name(state)]
    assert state_cookie.value == "/admin/"


def test_view_login_next(client, state):
    resp = client.get("/admin/login/", {"next": "/admin/page/42"})
    assert resp.status_code == 302
    redir = "https%3A%2F%2Ftestserver%2Fadmin%2Flogin%2F"
    assert (
        resp.headers["location"]
        == f"https://gandi.v5/authorize?client_id=Mei&redirect_uri={redir}&response_type=code&state={state}"
    )
    state_cookie = resp.cookies[get_cookie_name(state)]
    assert state_cookie.value == "/admin/page/42"


def test_view_login_next_is_login(client, state):
    resp = client.get("/admin/login/", {"next": "/admin/login/"})
    assert resp.status_code == 302
    redir = "https%3A%2F%2Ftestserver%2Fadmin%2Flogin%2F"
    assert (
        resp.headers["location"]
        == f"https://gandi.v5/authorize?client_id=Mei&redirect_uri={redir}&response_type=code&state={state}"
    )
    state_cookie = resp.cookies[get_cookie_name(state)]
    assert state_cookie.value == "/admin/"


def test_view_login_with_auth_code(client, state, auth_code, user_mei):
    client.cookies.load({get_cookie_name(state): "/admin/"})
    resp = client.get("/admin/login/", {"code": auth_code, "state": state})
    assert resp.status_code == 302
    assert resp.headers["Location"] == "/admin/"
    assert client.session['_auth_user_id'] == str(user_mei.pk)


def test_view_logout(client, user_mei):
    client.force_login(user_mei)
    assert '_auth_user_id' in client.session
    resp = client.get("/admin/logout/")
    assert resp.status_code == 302
    assert resp.headers["Location"] == "https://gandi.v5/logout"
    assert '_auth_user_id' not in client.session



import os
import os.path
from typing import Dict, Any

from django.core.exceptions import PermissionDenied

USERS = {
    "mey_accesstoken": {
        "username": "mei",
        "is_superuser": True,
    }
}


def load_userinfo(access_token):
    """
    Load userinfo from access token.

    Return a dict containing username, email, first_name, last_name, and
    is_superuser fields to register any new user.
    """
    try:
        return USERS[access_token]
    except KeyError:
        raise PermissionDenied


OAUTH2_VERIFY_CERTIFICATE = False
OAUTH2_TIMEOUT = 30

OAUTH2_LOAD_USERINFO = load_userinfo

OAUTH2_CLIENT_ID = "Mei"
OAUTH2_CLIENT_SECRET = "T0t0r0"

OAUTH2_AUTH_URL = "https://gandi.v5/authorize"
OAUTH2_TOKEN_URL = "https://gandi.v5/token"
OAUTH2_LOGOUT_URL = "https://gandi.v5/logout"


ROOT_URLCONF = "wagtail_oauth2.tests.urls"

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
SECRET_KEY = "beep"

DATABASES: Dict[str, Dict[str, Any]] = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(BASE_DIR, "test.db"),
        "TEST": {
            "NAME": "test.db",
        },
    },
}


INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "wagtail_oauth2",
    "wagtail.admin",
    "wagtail.users",
    "wagtail.core",
    "tests",
]

MIDDLEWARE = ["django.contrib.sessions.middleware.SessionMiddleware"]


TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
    }
]



"""
Views to log in using an OAuth2.0 Authorization Server.

Currently, we do not keep the access and refresh token in the session,
because it is not necessary.

It could be done later to retrieve data from API using OAuth2 authorizations.
"""
import binascii
import logging
import os

from django.conf import settings
from django.contrib.auth import get_user_model, login as auth_login
from django.contrib.auth.models import Group
from django.shortcuts import redirect
from django.urls import reverse
from wagtail.admin.views.account import LoginView, LogoutView

from .resources import Token
from .settings import get_setting


log = logging.getLogger(__name__)


def gen_state_name():
    """Generate a random state name."""
    return binascii.hexlify(os.urandom(5)).decode("ascii")


def get_cookie_name(state):
    """Generate the cookie name for the OAuth2.0 state."""
    return f"oauth.{state}"


class StateError(ValueError):
    """raised in case the oauth2 state workflow is not valid."""


def get_user_from_userinfo(userinfo):
    """Create or retrieve a user from the wagtail point of view, from the userinfo."""
    username = userinfo["username"]
    user_cls = get_user_model()
    try:
        user = user_cls.objects.get(username=username)
    except user_cls.DoesNotExist:
        # Create a new user. There's no need to set a password
        # because only the password from settings.py is checked.
        user = user_cls(username=username)

        user.is_staff = userinfo.get("is_staff", True)
        user.email = userinfo.get("email")
        user.first_name = userinfo.get("first_name")
        user.last_name = userinfo.get("last_name")
        user.is_superuser = userinfo["is_superuser"]
        user.save()
        if not user.is_superuser:
            groups = userinfo.get("groups", ["Moderators", "Editors"])
            for group_name in groups:
                group = Group.objects.filter(name=group_name).first()
                group.user_set.add(user)
                group.save()
    return user


class Oauth2LoginView(LoginView):
    """Login view."""

    template_name = "login_error.html"

    def add_state(self, request, resp, state):
        """Set the OAuth2.0 state in the cookie."""
        default = "/admin"
        if getattr(settings, "WAGTAIL_APPEND_SLASH", True):
            default += "/"
        referrer = request.GET.get("next", self.get_success_url())
        # never use the login form itself as referrer
        if referrer.startswith(reverse("wagtailadmin_login")):
            referrer = default
        if not referrer.startswith("/"):
            referrer = default

        host = request.get_host()
        resp.set_cookie(
            get_cookie_name(state),
            referrer,
            max_age=300,
            secure=not host.startswith("localhost:"),
            httponly=True,
        )
        return resp

    def check_state(self, request):
        """Verify that the OAuth2.0 state in the cookie match callback query."""
        state = request.GET.get("state")
        if not state:
            raise StateError("Missing OAuth2 state")
        cookie_name = get_cookie_name(state)
        if cookie_name not in request.COOKIES:
            raise StateError("Invalid OAuth2 state")
        return request.COOKIES[cookie_name]

    def start_oauth2_dance(self, request, state):
        """Create the http query that redirect to the OAuth2.0"""
        log.info("Redirect to the oauth2 authorization server")
        host = request.get_host()
        # XXX request.scheme does not works properly, maybe a config issue
        scheme = "http" if host.startswith("localhost:") else "https"
        url = Token.get_authenticated_url(
            scheme + "://" + host + reverse("wagtailadmin_login"),
            state,
        )
        return redirect(url)

    def consume_oauth2_code(self, request, redirect_uri):
        """Retrieve a bearer token from the authorization code, log the user."""
        log.info("Consume the code")
        token = Token.by_authcode(request.GET["code"])

        userinfo = get_setting("LOAD_USERINFO")(token["access_token"])
        user = get_user_from_userinfo(userinfo)
        auth_login(self.request, user)
        return redirect(redirect_uri)

    def render_error(self, error, error_description):
        """Render the template to diplay error to the user."""
        return self.render_to_response(
            {
                "error": error,
                "error_description": error_description,
            }
        )

    def get(self, request, *args, **kwargs):
        """Handle HTTP GET query."""
        redirect_uri = ""
        if "code" in request.GET or "error" in request.GET:
            try:
                redirect_uri = self.check_state(request)
            except StateError as exc:
                return self.render_error("OAuth 2 Security Error", str(exc))

        if "code" in request.GET:
            return self.consume_oauth2_code(request, redirect_uri)
        elif "error" in request.GET:
            log.info("OAuth2 server error")
            return self.render_error(
                request.GET["error"],
                request.GET.get("error_description"),
            )
        else:
            state = gen_state_name()
            return self.add_state(
                request, self.start_oauth2_dance(request, state), state
            )


class Oauth2LogoutView(LogoutView):
    """Logout view."""

    def dispatch(self, request, *args, **kwargs):
        """Handle every HTTP Query to logout user."""

        logout_url = get_setting("LOGOUT_URL")
        if logout_url:
            resp = redirect(logout_url)
        else:
            resp = super().dispatch(request, *args, **kwargs)

        # XXX sthis code code has been copy from Wagtail.

        # By default, logging out will generate a fresh sessionid cookie. We want to use the
        # absence of sessionid as an indication that front-end pages are being viewed by a
        # non-logged-in user and are therefore cacheable, so we forcibly delete the cookie here.
        resp.delete_cookie(
            settings.SESSION_COOKIE_NAME,
            domain=settings.SESSION_COOKIE_DOMAIN,
            path=settings.SESSION_COOKIE_PATH,
        )

        # HACK: pretend that the session hasn't been modified, so that SessionMiddleware
        # won't override the above and write a new cookie.
        self.request.session.modified = False
        return resp



"""Load settings for the wagtail-oauth2 app."""
from django.conf import settings

global_prefix = "OAUTH2_"


def get_setting(name, default=None):
    """Get the settings without the prefix."""
    return getattr(settings, global_prefix + name, default)



"""Override wagtail views to use Gandi ID as the login provider."""
from django.conf.urls import url

from .views import Oauth2LoginView, Oauth2LogoutView


urlpatterns = [
    url(r"^login/$", Oauth2LoginView.as_view(), name="wagtailadmin_login"),
    url(r"^logout/$", Oauth2LogoutView.as_view(), name="wagtailadmin_logout"),
]
"""Url to load in the app."""


# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['wagtail_oauth2', 'wagtail_oauth2.tests']

package_data = \
{'': ['*'], 'wagtail_oauth2': ['templates/*']}

install_requires = \
['Django>=3.2.4,<4.0.0', 'requests>=2.26.0,<3.0.0', 'wagtail>=2.14.1,<3.0.0']

setup_kwargs = {
    'name': 'wagtail-oauth2',
    'version': '0.2.0',
    'description': 'OAuth2.0 authentication fo wagtail',
    'long_description': 'Wagtail OAuth2.0 Login\n======================\n\n.. image:: https://readthedocs.org/projects/aioli-client/badge/?version=latest\n   :target: https://aioli-client.readthedocs.io/en/latest/?badge=latest\n   :alt: Documentation Status\n\n.. image:: https://github.com/Gandi/wagtail-oauth2/actions/workflows/main.yml/badge.svg\n   :target: https://github.com/Gandi/wagtail-oauth2/actions/workflows/main.yml\n   :alt: Build Status\n\n\n.. image:: https://codecov.io/gh/Gandi/wagtail-oauth2/branch/main/graph/badge.svg?token=VN14GVV3Y0\n   :target: https://codecov.io/gh/Gandi/wagtail-oauth2\n   :alt: Coverage\n\nPlugin to replace Wagtail default login by an OAuth2.0 Authorization Server.\n',
    'author': 'Guillaume Gauvrit',
    'author_email': 'guillaume@gandi.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://wagtail-oauth2.readthedocs.io/en/latest/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from django.urls import include, path

from wagtail.admin import urls as wagtailadmin_urls
from wagtail_oauth2 import urls as oauth2_urls

urlpatterns = [
    # /!\ must appears before admin/ to override the login part
    path("admin/", include(oauth2_urls)),
    path("admin/", include(wagtailadmin_urls)),
]
